import textwrap
from json.decoder import JSONDecodeError
from typing import Dict, Iterator

from yaml import YAMLError

from .jsonutil import loadjson_all, dumpjson
from .nodes import Constant, Empty, Node
from .parser import Parser, Environment
from .util import treat
from .value import Value
from .yamlutil import dumpyaml, loadyaml_all


def _evaluate(node: Node, *, args: Dict) -> Value:
    return node.evaluate(env=Environment(**args))


def _parse(doc: Value) -> Node:
    if not isinstance(doc, (list, dict)):
        return Constant(parent=Empty(), value=doc)
    return Parser().parse_node(doc, parent=Empty())


def _parse_json(s: str) -> Iterator[Node]:
    docs = loadjson_all(s)
    for doc in docs:
        yield _parse(doc)


def _parse_string(s: str) -> Iterator[Node]:
    s = textwrap.dedent(s)
    docs = loadyaml_all(s)
    count = 0
    try:
        for doc in docs:
            count += 1
            yield _parse(doc)
    except YAMLError as err:
        if count < 2:
            try:
                skip = 0
                it = _parse_json(s)
                while skip < count:
                    skip += 1
                    next(it)
                    continue
                yield from it
            except JSONDecodeError:
                raise err
        else:
            raise err


# noinspection PyShadowingBuiltins
def _parse_file(path: str, *, _open) -> Iterator[Node]:
    with _open(path) as f:
        docs = loadyaml_all(f)
        for doc in docs:
            yield _parse(doc)


def _render(node: Node, *, args: Dict, as_json: bool) -> str:
    value = _evaluate(node, args=args)
    if as_json:
        return dumpjson(value)
    else:
        return dumpyaml(value)


def render_string(s: str, *, args: Dict = None, as_json: bool = False) -> Iterator[str]:
    """Render each document from a string and return each rendered string one by one."""
    if not args:
        args = {}
    for node in _parse_string(s):
        yield _render(node, args=args, as_json=as_json)


def render_file(path: str, *, args: Dict = None, as_json: bool = False, _open=open) -> Iterator[str]:
    """Render each document from a file and return each rendered string one by one."""
    if not args:
        args = {}
    for node in _parse_file(path, _open=_open):
        yield _render(node, args=args, as_json=as_json)


def _render1(it: Iterator[str], as_json: bool) -> str:
    r = []
    if as_json:
        for doc in it:
            r.append(doc)
            r.append('\n')
    else:
        r.append(next(it))
        r.append("\n")
        for doc in it:
            r.append("\n--\n")
            r.append(doc)
    return "".join(r)


def render1s(s: str, *, args: Dict = None, as_json: bool = False) -> str:
    """Load all documents from a string and render them as string."""
    return _render1(render_string(s, args=args, as_json=as_json), as_json=as_json)


def render1f(path: str, *, args: Dict = None, as_json: bool = False) -> str:
    """Load all documents from a file and render them as string."""
    return _render1(render_file(path, args=args, as_json=as_json), as_json=as_json)


def load_string(s: str, *, args: Dict = None, numtype: type = float) -> Iterator[Value]:
    """Load all documents from a string."""
    if not args:
        args = {}
    docs = loadyaml_all(textwrap.dedent(s))
    for doc in docs:
        node = _parse(doc)
        value = _evaluate(node, args=args)
        yield treat(value, numtype=numtype)


def load_file(path: str, *, args: Dict = None, numtype: type = float, _open=open) -> Iterator[Value]:
    """Load all documents from a path."""
    if not args:
        args = {}
    with _open(path) as file:
        docs = loadyaml_all(file)
        for doc in docs:
            node = _parse(doc)
            value = _evaluate(node, args=args)
            yield treat(value, numtype=numtype)


def load1s(s: str, *, args: Dict = None, numtype: type = float) -> Value:
    """Load a single document from a string."""
    r, = load_string(s, args=args, numtype=numtype)
    return r


def load1f(path: str, *, args: Dict = None, numtype: type = float) -> Value:
    """Load a single document from a file."""
    r, = load_file(path, args=args, numtype=numtype)
    return r



import unittest
from dataclasses import dataclass

from jinsi.jsonutil import dumpjson, loadjson_all


class X:
    def __init__(self, *items):
        self._items = items

    def items(self):
        count = 0
        for item in self._items:
            yield count, item
            count += 1

@dataclass
class Z:
    foo: int
    bar: str


class JsonDumpTests(unittest.TestCase):

    def test_jsondump(self):
        self.assertEqual(dumpjson(X(1, "abc", 3)), '{"0":1,"1":"abc","2":3}')

    def test_loadall(self):
        result = tuple(loadjson_all("{}{}[]"))
        expected = ({}, {}, [])
        self.assertEqual(expected, result)

    def test_loadall_lines(self):
        result = tuple(loadjson_all("{}\n{}\n[]\n"))
        expected = ({}, {}, [])
        self.assertEqual(expected, result)

    def test_loadall_leading_whitespace(self):
        result = tuple(loadjson_all("\n{}\n{}\n[]\n"))
        expected = ({}, {}, [])
        self.assertEqual(expected, result)

    def test_dataclass(self):
        with self.assertRaises(TypeError):
            dumpjson(Z(1, 'quux'))
        res = dumpjson(Z(1, 'quux'), encode_dict=True)
        self.assertEqual('{"foo":1,"bar":"quux"}', res)


if __name__ == '__main__':
    unittest.main()



import textwrap
import unittest

from jinsi.yamlutil import dumpyaml


class YamlDumpTest(unittest.TestCase):

    def test_dump_yaml(self):
        self.assertEqual(textwrap.dedent("""\
            xyz: 123
            abc: 789
        """), dumpyaml({
            "xyz": 123,
            "abc": 789,
        }))


if __name__ == '__main__':
    unittest.main()



from .__pkginfo__ import version as __version__
from .api import \
    load_file, \
    load_string, \
    load1f, \
    load1s, \
    render_file, \
    render_string, \
    render1f, \
    render1s
from .functions import Functions
from .jsonutil import loadjson, loadjson_all, dumpjson
from .util import cached_method, cached_function
from .yamlutil import loadyaml, loadyaml_all, dumpyaml

__all__ = [
    'load_file',
    'load_string',
    'load1f',
    'load1s',
    'render_file',
    'render_string',
    'render1f',
    'render1s',

    'Functions',

    'loadjson',
    'loadjson_all',
    'dumpjson',

    'loadyaml',
    'loadyaml_all',
    'dumpyaml',
]



import unittest

from jinsi.functions import Functions


class FunctionsTest(unittest.TestCase):
    def test_concat_list(self):
        self.assertEqual([1, 2, 3], Functions.concat([1, 2], [3]))

    def test_concat_string(self):
        self.assertEqual("123", Functions.concat("12", "3"))

    def test_drop_right_list(self):
        self.assertEqual([1, 2], Functions.drop_right(2, [1, 2, 3, 4]))

    def test_drop_list(self):
        self.assertEqual([3, 4], Functions.drop(2, [1, 2, 3, 4]))

    def test_drop_right_string(self):
        self.assertEqual("ab", Functions.drop_right(2, "abcd"))

    def test_drop_string(self):
        self.assertEqual("cd", Functions.drop(2, "abcd"))

    def test_take_right_list(self):
        self.assertEqual([3, 4], Functions.take_right(2, [1, 2, 3, 4]))

    def test_take_list(self):
        self.assertEqual([1, 2], Functions.take(2, [1, 2, 3, 4]))

    def test_take_right_string(self):
        self.assertEqual("cd", Functions.take_right(2, "abcd"))

    def test_take_string(self):
        self.assertEqual("ab", Functions.take(2, "abcd"))

    def test_pad_left(self):
        self.assertEqual("00123", Functions.pad_left("123", "0", "5"))

    def test_pad_right(self):
        self.assertEqual("12300", Functions.pad_right("123", "0", "5"))

    def test_replace(self):
        self.assertEqual("foo qux bar foo qux", Functions.str_replace("foo", "foo qux", "foo bar foo"))

    def test_deepmerge(self):
        def o(**kw):
            return kw

        self.assertEqual(o(a=1, b=2, c=3), Functions.deepmerge(o(a=1), o(b=2), o(c=3)))
        self.assertEqual(o(a=1, b=2, c=3), Functions.deepmerge(o(a=8), o(b=2, a=1), o(c=3)))
        self.assertEqual(o(a=1, b=o(c=7)), Functions.deepmerge(o(a=1, b=o(c=3)), o(b=o(c=7))))

    def test_flatten(self):
        self.assertEqual([1, 2, 3], Functions.flatten([[1], [2, 3]]))

    def test_deepflatten(self):
        self.assertEqual([1, 2, 3, 4], Functions.deepflatten([1], [[2], [[3, 4]]]))


if __name__ == '__main__':
    unittest.main()



import unittest

from .common import JinsiTestCase


class JinsiExamples(JinsiTestCase):

    def test_conditional_with_let(self):
        doc = """\
            ::let:
                a: 1
            ::when:
                ::get: a == 1
            ::then:
                foo: one
            ::else:
                bar: two
        """

        expected = {'foo': 'one'}

        self.check(expected, doc)

    def test_when_with_let(self):
        doc = """\
            ::when:
                ::let:
                    a: 1
                ::get: a == 1
            ::then:
                foo: one
            ::else:
                bar: two
        """

        expected = {'foo': 'one'}

        self.check(expected, doc)

    def test_case(self):
        doc = """\
            value:
                ::let:
                    x: 3
                ::case:
                    x == 1: one
                    x == 2: two
                    x == 3: three
                    x == 4: four
        """

        expected = {'value': 'three'}

        self.check(expected, doc)

    def test_case_default_ellipsis(self):
        doc = """\
            value:
                ::let:
                    x: 17
                ::case:
                    x == 1: one
                    x == 2: two
                    x == 3: three
                    x == 4: four
                    ...: more than four
        """

        expected = {'value': 'more than four'}

        self.check(expected, doc)

    def test_case_default_underscore(self):
        doc = """\
            value:
                ::let:
                    x: 17
                ::case:
                    x == 1: one
                    x == 2: two
                    x == 3: three
                    x == 4: four
                    _: more than four
        """

        expected = {'value': 'more than four'}

        self.check(expected, doc)

    def test_case_when_without_get(self):
        doc = """\
            value:
                ::let:
                    x: 1
                ::when: x == 1
                ::then: one
                ::else: two
        """

        expected = {'value': 'one'}

        self.check(expected, doc)


if __name__ == '__main__':
    unittest.main()



import os
import re
import sys
import textwrap

from jinsi import *

import jinsi


def print_version(*, _print=print):
    _print(f"jinsi {jinsi.__version__}")


def print_help(*, _print=print):
    _print(textwrap.dedent(f"""
        {sys.argv[0]} [-j] [args...]
    
        ...where each argument may be:
        
          - a filename
          - a dash (to denote reading from standard input)
          - a variable binding of the form "key=value"
    
        The following options are recognized:
        
          -j  --json      Format output as JSON lines
        
        Standalone options:
    
          -v  --version   Print version information
          -h  --help      Print this help screen
        
    """))


def main(*argv, _print=print, _open=open, _stdin=sys.stdin):
    args = []
    env = {}
    fmt_json = False
    if argv:
        args_it = iter(argv)
    else:
        args_it = iter(sys.argv)
        next(args_it)
    opt_parsing = True
    for arg in args_it:
        if arg == "--":
            opt_parsing = False
            continue
        if opt_parsing:
            if arg in ("help", "version") and not os.path.exists(arg):
                arg = f"--{arg}"
            if arg in ("-v", "-version", "--version"):
                print_version(_print=_print)
                return
            if arg in ("-h", "-help", "--help"):
                print_help(_print=_print)
                return
            if arg in ("-j", "-json", "--json"):
                fmt_json = True
                continue
            m = re.match(r"([^=]+)=(.*)", arg)
            if m:
                key = m.group(1)
                val = m.group(2)
                env[key] = val
                continue
        args.append(arg)
    if not args:
        args = ["-"]
    count = 0
    for arg in args:
        if arg == '-':
            docs = render_string(_stdin.read(), args=env, as_json=fmt_json)
        else:
            docs = render_file(arg, args=env, as_json=fmt_json, _open=_open)
        for doc in docs:
            count += 1
            if fmt_json:
                _print(doc)
            else:
                if count > 1:
                    _print("---")
                if doc[-1] == '\n':
                    _print(doc, end='')
                else:
                    _print(doc)
